@@ -2,8 +2,9 @@ import json
 from dataclasses import dataclass
 from datetime import date, datetime
 from decimal import Decimal
+from functools import reduce
 from typing import List, Dict, Any
-from data_api_mapper.converters import GRAPHQL_CONVERTERS
+from data_api_mapper.converters import POSTGRES_PYTHON_MAPPER
 
 
 class ParameterBuilder:
@@ -18,6 +19,12 @@ class ParameterBuilder:
         else:
             return {'name': name, 'value': {type: value}}
 
+    @staticmethod
+    def json_serial(obj):
+        if isinstance(obj, date):
+            return obj.isoformat()
+        raise TypeError("Type %s not serializable" % type(obj))
+
     def add(self, name, value):
         if isinstance(value, str):
             self.result.append(self.build_entry_map(name, value, 'stringValue'))
@@ -29,10 +36,10 @@ class ParameterBuilder:
             self.result.append(self.build_entry_map(name, value, 'longValue'))
             return self
         elif isinstance(value, dict):
-            self.result.append(self.build_entry_map(name, json.dumps(value), 'stringValue', 'JSON'))
+            self.result.append(self.build_entry_map(name, json.dumps(value, default=self.json_serial), 'stringValue', 'JSON'))
             return self
         elif isinstance(value, list):
-            self.result.append(self.build_entry_map(name, json.dumps(value), 'stringValue', 'JSON'))
+            self.result.append(self.build_entry_map(name, json.dumps(value, default=self.json_serial), 'stringValue', 'JSON'))
             return self
         elif isinstance(value, float):
             self.result.append(self.build_entry_map(name, value, 'doubleValue'))
@@ -46,16 +53,11 @@ class ParameterBuilder:
         elif isinstance(value, Decimal):
             self.result.append(self.build_entry_map(name, str(value), 'stringValue', 'DECIMAL'))
             return self
-        else:
-            raise ValueError('The data type of the value does not match against any of the expected')
-
-    def add_or_null(self, name, value):
-        if value is None:
+        elif value is None:
             self.result.append(self.build_entry_map(name, True, 'isNull'))
             return self
         else:
-            self.add(name, value)
-            return self
+            raise ValueError('The data type of the value does not match against any of the expected')
 
     def add_dictionary(self, a_dict):
         for x in a_dict.keys():
@@ -65,6 +67,21 @@ class ParameterBuilder:
     def build(self):
         return self.result
 
+    def from_query(self, parameters) -> List:
+        if not parameters:
+            return []
+        if isinstance(parameters, list):
+            for param in parameters:
+                name = param['name']
+                value = param['value']
+                if 'cast' in param:
+                    self.result.append({'name': name, 'value': {'stringValue': value}, 'typeHint': param['cast']})
+                else:
+                    self.add(name, value)
+        elif isinstance(parameters, dict):
+            self.add_dictionary(parameters)
+        return self.build()
+
 
 @dataclass
 class RowMetadata:
@@ -82,16 +99,8 @@ class RowMetadata:
 class QueryMetadata:
     rows: List[RowMetadata]
 
-    @property
-    def main_table(self):
-        names = [x.table_name for x in self.rows]
-        return max(names, key=names.count)
-
     def field_names(self):
-        main_table = self.main_table
-        return [x.name if x.table_name == main_table or not x.table_name
-                else f'{x.table_name}_{x.name}'
-                for x in self.rows]
+        return [x.name for x in self.rows]
 
     def converters(self, converter_map) -> List:
         return [converter_map.get(x.type_name, None) for x in self.rows]
@@ -109,23 +118,23 @@ class QueryResponse:
 
 
 class Transaction:
-    def __init__(self, rds_client, secret_arn, cluster_arn, database_name) -> None:
+    def __init__(self, data_client) -> None:
         super().__init__()
-        self.rds_client = rds_client
-        self.secret_arn = secret_arn
-        self.cluster_arn = cluster_arn
-        self.database_name = database_name
-        transaction = rds_client.begin_transaction(
+        self.rds_client = data_client.rds_client
+        self.secret_arn = data_client.secret_arn
+        self.cluster_arn = data_client.cluster_arn
+        self.database_name = data_client.database_name
+        self.data_client = data_client
+        transaction = self.rds_client.begin_transaction(
             secretArn=self.secret_arn, database=self.database_name, resourceArn=self.cluster_arn
         )
         self.transaction_id = transaction['transactionId']
 
-    def execute(self, sql, parameters=()) -> Dict[str, Any]:
-        return self.rds_client.execute_statement(
-            secretArn=self.secret_arn, database=self.database_name,
-            resourceArn=self.cluster_arn, includeResultMetadata=True,
-            sql=sql, parameters=parameters, transactionId=self.transaction_id
-        )
+    def query(self, sql, parameters=(), mapper=POSTGRES_PYTHON_MAPPER) -> Dict[str, Any]:
+        return self.data_client.query(sql, parameters, mapper, self.transaction_id)
+
+    def batch_query(self, sql, parameters=()) -> Dict[str, Any]:
+        return self.data_client.batch_query(sql, parameters, self.transaction_id)
 
     def commit(self) -> Dict[str, str]:
         return self.rds_client.commit_transaction(
@@ -140,23 +149,64 @@ class Transaction:
 
 class DataAPIClient:
 
-    def __init__(self, rds_client, secret_arn, cluster_arn, database_name) -> None:
+    def __init__(self, rds_client, secret_arn, cluster_arn, database_name, mapper=POSTGRES_PYTHON_MAPPER) -> None:
         super().__init__()
         self.rds_client = rds_client
         self.secret_arn = secret_arn
         self.cluster_arn = cluster_arn
         self.database_name = database_name
+        self.mapper = mapper
 
-    def execute(self, sql, parameters=(), wrap_result=True) -> QueryResponse:
-        response = self.rds_client.execute_statement(
-            secretArn=self.secret_arn, database=self.database_name,
-            resourceArn=self.cluster_arn, includeResultMetadata=True,
-            sql=sql, parameters=parameters
-        )
-        return QueryResponse.from_dict(response) if wrap_result else response
+    def query(self, sql, parameters=None, mapper=None, transaction_id=None):
+        this_mapper = mapper if mapper is not None else self.mapper
+        data_client_params = ParameterBuilder().from_query(parameters)
+        config = {
+            'secretArn': self.secret_arn, 'database': self.database_name,
+            'resourceArn': self.cluster_arn, 'includeResultMetadata': True,
+            'sql': sql, 'parameters': data_client_params
+        }
+        if transaction_id is not None:
+            config['transactionId'] = transaction_id
+        response = self.rds_client.execute_statement(**config)
+        if 'columnMetadata' in response:
+            response = QueryResponse.from_dict(response)
+            return DictionaryMapper(response.metadata, this_mapper).map(response.records)
+        else:
+            return response['numberOfRecordsUpdated']
+
+    def batch_query(self, sql, parameter_list=(), transaction_id=None):
+        data_client_params = [ParameterBuilder().from_query(x) for x in parameter_list]
+        config = {
+            'secretArn': self.secret_arn, 'database': self.database_name, 'resourceArn': self.cluster_arn,
+            'sql': sql, 'parameterSets': data_client_params
+        }
+        if transaction_id is not None:
+            config['transactionId'] = transaction_id
+        response = self.rds_client.batch_execute_statement(**config)
+        return response
+
+    def query_paginated(self, sql, parameters=None, mapper=POSTGRES_PYTHON_MAPPER, page_size=100):
+        paginator = self.paginator(sql, parameters, mapper, page_size)
+        return reduce(lambda x, y: x+y, paginator)
+
+    def paginator(self, sql, parameters=None, mapper=POSTGRES_PYTHON_MAPPER, page_size=100):
+        offset = 0
+
+        def paginate():
+            nonlocal offset
+            while True:
+                sql_paginated = f'{sql} limit {page_size} offset {offset}'
+                response = self.query(sql_paginated, parameters, mapper)
+                yield response
+                if len(response) < page_size:
+                    return
+                else:
+                    offset += page_size
+
+        return paginate()
 
     def begin_transaction(self):
-        return Transaction(self.rds_client, self.secret_arn, self.cluster_arn, self.database_name)
+        return Transaction(self)
 
 
 class DictionaryMapper:
@@ -175,9 +225,3 @@ class DictionaryMapper:
 
     def map(self, records):
         return [self.map_record(x) for x in records]
-
-
-class GraphQLMapper(DictionaryMapper):
-
-    def __init__(self, metadata: QueryMetadata):
-        super().__init__(metadata, GRAPHQL_CONVERTERS)
